local AddonName, AddonTable = ...

local Addon = AddonTable.Addon

local registry = LibStub:GetLibrary("CallbackHandler-1.0"):New(Addon)
local usedEvents = 0
LibStub("AceEvent-3.0"):Embed(registry)
LibStub("AceBucket-3.0"):Embed(registry)

local vigilanceId = 50720
local vigilanceName = GetSpellInfo(vigilanceId)


local unknownUnit = {}

unknownUnit.name = UNKNOWNOBJECT
unknownUnit.GUID = UnitGUID("none")

function unknownUnit:SetTarget(target)
  return
end

function unknownUnit:RemoveTarget()
  return
end


local unitPrototype = {}

function unitPrototype:HasVigilance()
  return self.hasVigilance ~= nil
end

function unitPrototype:GetExpirationTime()
  return self.expirationTime
end

function unitPrototype:GetOwner()
  return self.owner
end

function unitPrototype:SetVigilance(owner, expirationTime)
  if self.hasVigilance then
    if self.owner == owner then
      -- Check if buff was recast
      if self.expirationTime ~= expirationTime then
        self.expirationTime = expirationTime
        registry:Fire("AURA_REFRESHED", self)
      end
      return
    elseif owner == unknownUnit then
      -- Do nothing, assume it's the same
      return
    elseif self.owner ~= unknownUnit then
      self:RemoveVigilance()
    end
  end
  owner:SetTarget(self)
  self.hasVigilance = true
  self.expirationTime = expirationTime
  self.owner = owner
  registry:Fire("AURA_GAINED", self)
end

function unitPrototype:RemoveVigilance()
  if not self.hasVigilance then return end
  self.owner:RemoveTarget(self)
  self.hasVigilance = nil
  self.expirationTime = nil
  self.owner = nil
  registry:Fire("AURA_LOST", self)
end

function unitPrototype:SetTarget(target)
  if self.target then
    self.target:RemoveVigilance()
  end
  self.target = target
end

function unitPrototype:GetTarget()
  return self.target
end

function unitPrototype:RemoveTarget()
  self.target = nil
end


local unitMetatable = { __index = unitPrototype }

local units = {}

local function GetUnitByID(unitID)
  if unitID then
    local unitGUID = UnitGUID(unitID)
    local unit = units[unitGUID]
    if not unit then
      unit = setmetatable({ name = UnitName(unitID), GUID = unitGUID }, unitMetatable)
      units[unitGUID] = unit
    end
    return unit
  else
    return unknownUnit
  end
end

local function FilterUnits(predicate)
  for GUID, unit in pairs(units) do
    if not predicate(unit) then
      units[GUID] = nil
    end
  end
end


local function RequiredFilter(unit)
  return unit:HasVigilance() or unit:GetTarget() ~= nil
end

local function GroupFilter(unit)
  return UnitExists(unit.name) ~= nil and (unit:HasVigilance() or unit:GetTarget() ~= nil)
end

local function NoneFilter(unit)
  return false
end


local function CheckUnit(targetID)
  if not UnitExists(targetID) then return end

  local name, _, _, _, _, _, expirationTime, ownerID, _ = UnitBuff(targetID, vigilanceName)

  if name ~= nil then -- Buff found
    local target = GetUnitByID(targetID)
    local owner = GetUnitByID(ownerID)
    target:SetVigilance(owner, expirationTime)
  else
    local unit = units[UnitGUID(targetID)]
    if unit ~= nil then
      unit:RemoveVigilance()
    end
  end
end

local function ScanRoster()
  if GetNumRaidMembers() > 0 then
    for i = 1, 40 do
      CheckUnit("raid"..i)
      CheckUnit("raidpet"..i)
    end
  else
    CheckUnit("player")
    CheckUnit("playerpet")
    for i = 1, 4 do
      CheckUnit("party"..i)
      CheckUnit("partypet"..i)
    end
  end
  FilterUnits(GroupFilter)
end

local function OnUnitAura(unitIDs)
  for unitID in pairs(unitIDs) do
    CheckUnit(unitID)
  end
  FilterUnits(RequiredFilter)
end

local nop = function () end
local fire = registry.Fire
registry.Fire = nop

function registry:OnUsed()
  if usedEvents == 0 then
    registry:RegisterEvent("PARTY_MEMBERS_CHANGED", ScanRoster)
    registry:RegisterEvent("RAID_ROSTER_UPDATE", ScanRoster)
    registry:RegisterBucketEvent("UNIT_AURA", 0.63, OnUnitAura)

    ScanRoster()
    registry.Fire = fire
  end
  usedEvents = usedEvents + 1
end

function registry:OnUnused()
  usedEvents = usedEvents - 1
  if usedEvents == 0 then
    registry.Fire = nop

    registry:UnregisterAllEvents()
    registry:UnregisterAllBuckets()

    FilterUnits(NoneFilter)
  end
end

function Addon:IterateTargets()
  local lastIndex = nil
  local unit
  return function ()
    repeat
      lastIndex, unit = next(units, lastIndex)
    until unit == nil or unit:HasVigilance()
    return unit
  end
end
